from cloudmesh.common.Shell import Shell
from cloudmesh.common.console import Console
import os
from cloudmesh.common.FlatDict import FlatDict
from textwrap import dedent
import yaml
from cloudmesh.common.util import banner
from cloudmesh.common.StopWatch import StopWatch
import socket

class Rivanna:

    def jupyter(self, port=8000):
        """
        Start a Jupyter notebook on Rivanna.

        Args:
            port (int): The port number for Jupyter. Default is 8000.

        Note:
            This method assumes a VPN connection is active.
        """
        self.port = port

        # test if vpn is on

        # start login on machine and start jupyter
        "jupyter notebook --no-browser --port=<PORT>"
        # open tunnel
        "ssh -L 8080:localhost:<PORT> <REMOTE_USER>@<REMOTE_HOST>"
        # access to notebook on localhost

    def __init__(self, host="rivanna", debug=False):
        """
        Initialize the Rivanna class.

        Args:
            host (str): The hostname, default is "rivanna".
            debug (bool): Enable debugging if True, default is False.
        """
        self.debug = debug
        self.data = dedent(
          """
          macos:
            macos:
              partition: "macos"
              account: "macos"
              gres": "none"
          rivanna:
            parallel:
              partition: "parallel"
              account: "bii_dsc_community"
              nodes: 2
              ntask-per-node: 4
            v100:
              gres: "gpu:v100:1"
              partition: "bii-gpu"
              account: "bii_dsc_community"
            a100:
              gres: "gpu:a100:1"
              partition: "gpu"
              account: "bii_dsc_community"
            a100-dgx:
              gres: "gpu:a100:1"
              reservation: "bi_fox_dgx"
              partition: "bii-gpu"
              account: "bii_dsc_community"
            k80:
              gres: "gpu:k80:1"
              partition: "gpu"
              account: "bii_dsc_community"
            p100:
              gres: "gpu:p100:1"
              partition: "gpu"
              account: "bii_dsc_community"
            a100-pod:
              gres: "gpu:a100:1"
              account: "bii_dsc_community"
              constraint: "gpupod"
              partition: gpu
            rtx2080:
              gres: "gpu:rtx2080:1"
              partition: "gpu"
              account: "bii_dsc_community"
            rtx3090:
              gres: "gpu:rtx3090:1"
              partition: "gpu"
              account: "bii_dsc_community"          
          greene:
            v100:
              gres: "gpu:v100:1"
            a100:
              gres: "gpu:a100:1"
          maltlab:
            rtx_titan:
              gres: "gpu:rtx_titan:1"
        """
        )
        self.directive = yaml.safe_load(self.data)

    def parse_sbatch_parameter(self, parameters):
        """
        Parse the parameters string and convert it to a dictionary.

        Args:
            parameters (str): Comma-separated string of parameters.

        Returns:
            dict: A dictionary containing parsed parameters.
        """
        result = {}
        data = parameters.split(",")
        for line in data:
            key, value = line.split(":",1)
            result[key] = value
        return result

    def directive_from_key(self, key):
        """
        Retrieve the Slurm directives for a specific key.

        Args:
            key (str): The key to retrieve directives.

        Returns:
            dict: A dictionary containing Slurm directives for the specified key.
        """
        return self.directive[key]

    def create_slurm_directives(self, host=None, key=None):
        """
        Create Slurm directives based on the provided host and key.

        Args:
            host (str): The hostname.
            key (str): The key to retrieve directives.

        Returns:
            str: A string containing Slurm directives for the specified host and key.
        """
        directives = self.directive[host][key]
        block = ""

        def create_direcitve(name):
            return f"#SBATCH --{name}={directives[name]}\n"

        for key in directives:
            block = block + create_direcitve(key)

        return block


    def login(self, host, key):
        """
        SSH on Rivanna by executing an interactive job command.

        Args:
            host (str): The hostname.
            key (str): The key used for authentication.

        Returns:
            str: Empty string.
        """

        def create_parameters(host, key):

            directives = self.directive[host][key]
            block = ""

            def create_direcitve(name):
                return f" --{name}={directives[name]}"

            for key in directives:
                block = block + create_direcitve(key)

            return block


        parameters = create_parameters(host, key)
        command = f'ssh -tt {host} "/opt/rci/bin/ijob{parameters}"'

        Console.msg(command)
        if not self.debug:
             os.system(command)
        return ""


    def cancel(self, job_id):
        """
        Cancel the job with the given ID.

        Args:
            job_id (str): The ID of the job to be canceled.

        Raises:
            NotImplementedError: This method is not implemented.
        """
        raise NotImplementedError

    def storage(self, directory=None):
        """
        Get information about the specified directory.

        Args:
            directory (str): The directory to get information about.

        Raises:
            NotImplementedError: This method is not implemented.
        """
        raise NotImplementedError

    def edit(self, filename=None, editor="emacs"):
        """
        Start the command-line editor of choice on the file on Rivanna in the current terminal.

        Args:
            filename (str): The name of the file to be edited.
            editor (str): The command-line editor to use, default is "emacs".

        Returns:
            None
        """
        raise NotImplementedError

    def create_apptainer_image(self, name):
        """
        Create a apptainer image on Rivanna.

        Args:
            name (str): The name of the apptainer image.

        Returns:
            None

        Requires:
            export APPTAINER_CACHEDIR=/scratch/$USER/.apptainer/
            or
            export APPTAINER_CACHEDIR=/$HOME/.apptainer/

            Please note it is prefered to use scratch as the home directory may have too little storage

        """

        try:
            cache = os.environ["APPTAINER_CACHEDIR"]
            banner("Cloudmesh Rivanna Singularity Build")

            image = os.path.basename(name.replace(".def", ".sif"))

            print("Image name       :", image)
            print("Singularity cache:", cache)
            print("Definition       :", name)
            print()
            StopWatch.start("build image")
            os.system("apptainer build {image} {name}")
            StopWatch.stop("build image")
            size = Shell.run(f"du -sh {image}").split()[0]
            timer = StopWatch.get("build image")
            print()
            print(f"Time to build {image}s ({size}) {timer}s")
            print()

        except Exception as e:
            Console.error(e, traceflag=True)
            




import shutil

import yaml

from cloudmesh.shell.command import command
from cloudmesh.shell.command import PluginCommand
from cloudmesh.rivanna.rivanna import Rivanna
from cloudmesh.common.console import Console
from cloudmesh.common.util import path_expand
from pprint import pprint
from cloudmesh.common.debug import VERBOSE
from cloudmesh.shell.command import map_parameters
from cloudmesh.common.parameter import Parameter
from cloudmesh.common.variables import Variables
from cloudmesh.common.util import banner
from cloudmesh.common.Shell import Shell

class RivannaCommand(PluginCommand):

    # noinspection PyUnusedLocal
    @command
    def do_rivanna(self, args, arguments):
        """
        ::

          Usage:
                rivanna storage info DIRECTORY [--info]
                rivanna login [--sbatch=SBATCH] [--host=HOST] [KEY] [--debug]
                rivanna slurm info
                rivanna slurm [--sbatch=SBATCH] [--host=HOST] [KEY] [--debug]
                rivanna tutorial [KEYWORD]
                rivanna vpn on
                rivanna vpn off
                rivanna vpn info
                rivanna vpn status
                rivanna ticket
                rivanna image build DEFFILE
                rivanna jupyter --port=PORT

          This command simplifys access to rivanna.

          Arguments:
              DIRECTORY  a location of a directory on rivanna

          Options:
              -f      specify the file

          Description:

            rivanna storage info DIRECTORY [--info]
                A command that can be executed remotely and obtains information about the storage associated
                with a directory on rivanna

            rivanna login [--allocation=ALLOCATION] [--gres=GRES] [--cores=CORES] [--host=HOST] [--constraint=CONSTRAINT] [--reservation=RESERVATION] [--time=TIME]
                A command that logs into from your current computer into an interactive node on rivanna
                with a given GPU. Values for GPU are

                a100 or a100-80 -- uses a A100 with 80GB
                a100-40 -- uses a A100 with 40GB
                a100-localscratch  -- uses a A100 with 80GB and access to localscratch

                others to be added from rivannas hardware description

            rivanna tutorial apptainer
                shows the rivanna apptainer documentation on the UVA website

            rivanna tutorial hpc
                shows the general rivanna hpc information on infomall.org

            rivanna tutorial pod
                shows the general rivanna pod information on infomall.org

            rivanna tutorial globus
                shows the general rivanna globus information on infomall.org

            rivanna tutorial rclone
                shows the general rivanna rclone information on infomall.org

            rivanna vpn on
                switches the vpn on

            rivanna vpn off
                switches the vpn off

            rivanna vpn info
                prints the information about the current connection to the
                internet. INformation includes

                  "ip": "128.143.1.11",
                  "hostname": "vpn-user-1-11.itc.virginia.edu",
                  "city": "Charlottesville",
                  "region": "Virginia",
                  "country": "US",
                  "loc": "38.0293,-78.4767",
                  "org": "AS225 University of Virginia",
                  "postal": "22902",
                  "timezone": "America/New_York",
                  "readme": "https://ipinfo.io/missingauth"

            rivanna vpn status
                prints True if vpn is enabled, False if not


          Installation:

            pip install cloudmesh-rivana
            cms help

        """


        # arguments.FILE = arguments['--file'] or None

        # switch debug on

        variables = Variables()
        variables["debug"] = True

        map_parameters(arguments,
                       "port",
                       "host",
                       "sbatch",
                       "debug")

        host = arguments.host = arguments.host or "rivanna"

        # VERBOSE(arguments)

        key = arguments.KEY
        # VERBOSE(arguments)


        rivanna = Rivanna()
        rivanna.debub = arguments.debug

        def VPN():
            from cloudmesh.vpn.vpn import Vpn
            vpn = Vpn(None,
                      timeout=30)
            return vpn

        if arguments.jupyter:

            port = arguments.port or 8000
            rivanna.jupyter(port=port)

        elif arguments.storage:

            Console.error("not implemented")

        elif arguments.image and arguments.build:

            buidlfile = arguments.DEFFILE

            rivanna.create_apptainer_image(buidlfile)

        elif arguments.login:

            key = arguments.KEY

            rivanna.directive[host][key].update({"time": "30:00"})

            if arguments.sbatch:

                data = rivanna.parse_sbatch_parameter(arguments.sbatch)
                rivanna.directive[host][key].update(data)

            d = rivanna.directive[host][key]

            time = arguments.time or "30:00"

            rivanna.login(host, key)

        elif arguments.slurm and arguments.info:

           print(yaml.dump(rivanna.directive))

        elif arguments.slurm and arguments.KEY and not arguments.login:

            key = arguments.KEY

            if arguments.sbatch:
                data = rivanna.parse_sbatch_parameter(arguments.sbatch)
                rivanna.directive[host][key].update(data)


            d = rivanna.directive[host][key]
            slurm_directive = rivanna.create_slurm_directives(host=host, key=key)

            print(slurm_directive)


        elif arguments.vpn and arguments.on:

            vpn = VPN()

            Console.ok("Connecting ... ")
            vpn.connect()
            if vpn.enabled():
                Console.ok("ok")
            else:
                Console.error("failed")

        elif arguments.vpn and arguments.off:

            vpn = VPN()

            Console.ok("Disconnecting ... ")
            vpn.disconnect()
            if not vpn.enabled():
                Console.ok("ok")
            else:
                Console.error("failed")

        elif arguments.vpn and arguments.status:

            vpn = VPN()

            print(vpn.enabled())

        elif arguments.vpn and arguments.info:

            vpn = VPN()

            print(vpn.info())

        elif arguments.tutorial:

            keyword = arguments.KEYWORD

            if keyword in ["pod"]:
                Shell.browser("https://infomall.org/uva/docs/tutorial/rivanna-superpod/")

            elif keyword in ["rclone"]:
                Shell.browser("https://infomall.org/uva/docs/tutorial/rclone/")

            elif keyword in ["globus"]:
                Shell.browser("https://infomall.org/uva/docs/tutorial/globus/")

            elif keyword in ["apptainer"]:
                #rivanna.browser("https://infomall.org/uva/docs/tutorial/singularity/")
                Shell.browser("https://www.rc.virginia.edu/userinfo/rivanna/software/apptainer/")

            elif keyword in ["training"]:
                Shell.browser("https://infomall.org/uva/docs/tutorial/cybertraining/")

            elif keyword in ["hpc", "system"]:
                Shell.browser("https://infomall.org/uva/docs/tutorial/rivanna/")

            else:
                Shell.browser("https://infomall.org/uva/docs/tutorial/")

        elif arguments.ticket:

            Shell.browser("https://www.rc.virginia.edu/form/support-request/")

        return ""
